////////////////////////////////////////////////////////////////////////////////
//                                                                            //
// The contents of this file are subject to the Mozilla Public License        //
// Version 1.1 (the "License"); you may not use this file except in           //
// compliance with the License. You may obtain a copy of the License at       //
// http://www.mozilla.org/MPL/                                                //
//                                                                            //
// Software distributed under the License is distributed on an "AS IS" basis, //
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License   //
// for the specific language governing rights and limitations under the       //
// License.                                                                   //
//                                                                            //
// The Original Code is RawToPsd.dpr, releases April 30, 2002.                //
//                                                                            //
// The Initial Developer of the Original Code is Dan Strandberg               //
// <dan@game-editing.net>. Portions created by Dan Strandberg are Copyright   //
// (C) 2001-2002 Game-Editing.net. All Rights Reserved.                       //
//                                                                            //
// Contributor(s):                                                            //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////
program RawToPsd;

{$APPTYPE CONSOLE}
{$WARN SYMBOL_PLATFORM OFF}

uses
  SysUtils, Classes, Windows;

////////////////////////////////////////////////////////////////////////////////
//                                                                            //
// Dungeon Siege Raw Format:                                                  //
//                                                                            //
//  [ HEADER ]                                                                //
//  [ MIP  0 ]                                                                //
//  [ MIP  1 ] <-- if surfacecount > 1                                        //
//  [ MIP  X ] <--        "                                                   //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//                                                                            //
// PSD Format                                                                 //
//                                                                            //
//  The full specs comes on 84 A4 pages so it's a wee bit simplified here :)  //
//                                                                            //
//  [ Header            ]                                                     //
//  [ Color mode data   ]                                                     //
//  [ Image resources   ]                                                     //
//  [ Layer & Mask Info ]                                                     //
//  [ Image Data        ]                                                     //
//                                                                            //
//  Color Mode Data starts with Length: LongWord; so just skip ahead that     //
//  many bytes if reading or set it to 0 if writing. Same goes for            //
//  "Image Resources" and "Layer & Mask Info".                                //
//                                                                            //
//  The ImageData starts with Compression: Word; 0 = RAW, 1 = RLE, 2 = Zip    //
//  w/o prediction, 3 = Zip w/ prediction. The actual pixels are stored one   //
//  channel after eachother. Like: RRRRRR... GGGGGG... BBBBBB... AAAAA... Etc //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

type
  FOURCC = Array[1..4] of char;

  TRawHeader = packed record
    HeaderMagic  : FOURCC; // special magic number
    Format       : FOURCC; // format of bits
    Flags        : Word;   // any special flags (for future expansion)
    SurfaceCount : Word;   // total surfaces stored (for mip maps), always >= 1
    Width        : Word;   // width of surface 0
    Height       : Word;   // height of surface 0
  end;

  TPsdHeader = packed record
    Signature : FOURCC;
    Version   : Word;
    Reserved  : array[0..5] of Byte;
    Channels  : Word;
    Rows      : Cardinal;
    Columns   : Cardinal;
    Depth     : Word;
    Mode      : Word;
  end;

const
  HEADER_MAGIC: FOURCC     = 'ipaR';
  FORMAT_ARGB_8888: FOURCC = '8888';
  FLAG_NONE                = $0000;

function ReverseLong(Value: LongWord): LongWord;
asm
  bswap EAX
  mov   ECX, EAX
end;

function ReverseShort(Value: Word): Word;
asm
  xchg  AL, AH
  mov   CX, AX
end;

procedure DisplayHeader;
begin
  WriteLn('RawToPsd 0.6');
  WriteLn('Copyright (C) 2002 Dan Strandberg');
  Writeln('');
  Writeln('');
end;

procedure DisplayInfo;
begin
  Writeln('  USAGE:');
  Writeln('    RawToPsd <file1.raw [file2.raw file3.. etc]>');
end;

procedure Convert(InFile, OutFile: string);
var
  Src, Dst: TFileStream;
  RawHdr: TRawHeader;
  x,y:integer;
  dp,sp:pbyte;
  d,s:pointer;
  PsdHeader: TPsdHeader;
begin
  dst := nil;
  src := TFileStream.Create(InFile, fmopenread);
  dst := TFileStream.Create(OutFile, fmCreate);
  try
    // Read header
    src.Read(RawHdr, SizeOf(TRawHeader));

    if (RawHdr.HeaderMagic <> HEADER_MAGIC) or (RawHdr.SurfaceCount < 1) then
    begin
      Writeln(InFile + ' is not a .raw file');
      Exit;
    end;
    if RawHdr.Format <> FORMAT_ARGB_8888 then
    begin
      Writeln('Unsupported Format');
      Exit;
    end;

    with PsdHeader do
    begin
      Signature := '8BPS';
      Version   := ReverseShort(1);
      FillChar(Reserved, 6, 0);
      Channels  := ReverseShort(4);
      Rows      := ReverseLong(RawHdr.Height);
      Columns   := ReverseLong(RawHdr.Width);
      Depth     := ReverseShort(8);
      Mode      := ReverseShort(3);
    end;
    Dst.Write(PsdHeader, SizeOf(TPsdHeader));

    x := 0;
    dst.Write(x, 4); // Color mode data length
    dst.Write(x, 4); // Image resources length
    dst.Write(x, 4); // Layer & Mask information length
    dst.Write(x, 2); // Compression method. RAW = 0

    GetMem(s, RawHdr.Width * RawHdr.Height * 4);
    try
      src.Read(s^, RawHdr.Width * RawHdr.Height * 4);
      GetMem(d, RawHdr.Width * RawHdr.Height * 4);
      try
        dp := d;
        // It could easilly be changed to one loop and maybe use a "dp" for
        // each channel, but who cares.

        // Red
        for y := RawHdr.Height -1 downto 0 do
        begin
          Cardinal(sp) := Cardinal(s) + y*RawHdr.Width*4 + 2;
          for x := 0 to RawHdr.Width - 1 do
          begin
            dp^ := sp^;
            Inc(dp);
            Inc(sp, 4);
          end;
        end;

        // Green
        for y := RawHdr.Height -1 downto 0 do
        begin
          Cardinal(sp) := Cardinal(s) + y*RawHdr.Width*4 + 1;
          for x := 0 to RawHdr.Width - 1 do
          begin
            dp^ := sp^;
            Inc(dp);
            Inc(sp, 4);
          end;
        end;

        // Blue
        for y := RawHdr.Height -1 downto 0 do
        begin
          Cardinal(sp) := Cardinal(s) + y*RawHdr.Width*4;
          for x := 0 to RawHdr.Width - 1 do
          begin
            dp^ := sp^;
            Inc(dp);
            Inc(sp, 4);
          end;
        end;

        // Alpha
        for y := RawHdr.Height -1 downto 0 do
        begin
          Cardinal(sp) := Cardinal(s) + y*RawHdr.Width*4 + 3;
          for x := 0 to RawHdr.Width - 1 do
          begin
            dp^ := sp^;
            Inc(dp);
            Inc(sp, 4);
          end;
        end;

        dst.Write(d^, RawHdr.Width * RawHdr.Height * 4);
      finally
        FreeMem(d);
      end;
    finally
      FreeMem(s);
    end;
  finally
    FreeAndNil(src);
    if dst <> nil then FreeAndNil(dst);
  end;
end;

// http://www.delphi3000.com/articles/article_299.asp
function GetLongPathName(Path : string) : string;
var
  HFile : THandle;
  i     : integer;
  Fd    : TWin32FindData;
begin
  Result := '';
  if (pos ('\\', Path) + pos ('*', Path) + pos ('?', Path) <> 0) then
  begin
    { ignore NetBIOS name, joker chars }
    Result := Path;
    exit;
  end;

  Path := ExpandFileName (Path);
  i := Length (Path);
  if i <= 0 then
    Exit;
  if (i > 3) and (Path[i] = '\') then
    SetLength (Path, i - 1);
  while (Length(Path) > 10) do
  begin
    HFile := FindFirstFile (PChar(Path), Fd);
    if (HFile = INVALID_HANDLE_VALUE) then
    begin
      Result := '';
      Exit;
    end;
    Windows.FindClose (HFile);
    Result := '\' + string (Fd.cFileName) + Result;
    Path := ExtractFilePath (Path);
    SetLength (Path, Length(Path)-1);
  end;
  Result := Path + Result;
end;


var
  i: integer;
  f: string;
begin
  DisplayHeader;
  if ParamCount < 1 then
  begin
    DisplayInfo;
    ReadLn;
  end
  else
  begin
    for i := 1 to ParamCount do
    try
      f := ParamStr(i);
      Convert(f, ChangeFileExt(GetLongPathName(f),'.psd'));
    except
      Writeln('Doh, something went wrong :)'); // <- STD Informative error! :)
    end;
  end;
end.
